/* stat2report.cc
****************************************************************************************************************
****************************************************************************************************************

    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2020, 2021 Askar
    Almatov

    This file is part of Trafficmeter.

    Trafficmeter is free software: you can redistribute it and/or modify it under the terms of the GNU General
    Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.

    Trafficmeter is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
    License for more details.

    You should have received a copy of the GNU General Public License along with Trafficmeter. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include <ctime>
#include <iomanip>
#include <iostream>

#include "stat2report.h"

using namespace trafficmeter;

using std::cout;
using std::endl;
using std::fixed;
using std::left;
using std::right;
using std::setprecision;
using std::setw;
using std::string;

/**************************************************************************************************************/
Stat2Report::Stat2Report() :
    shouldHideTime_( false ),
    shouldShowVlans_( false ),
    shouldShowHosts_( false ),
    shouldShowFlows_( false ),
    divisor_( 1.0 ),
    minimalTraffic_( 0.0 )
{
}

/**************************************************************************************************************/
void
Stat2Report::hideTime()
{
    shouldHideTime_ = true;
}

/**************************************************************************************************************/
void
Stat2Report::showVlans()
{
    shouldShowVlans_ = true;
}

/**************************************************************************************************************/
void
Stat2Report::showHosts()
{
    shouldShowHosts_ = true;
}

/**************************************************************************************************************/
void
Stat2Report::showFlows()
{
    shouldShowFlows_ = true;
}

/**************************************************************************************************************/
bool
Stat2Report::setDivisor( double value )
{
    if ( divisor_ != 1.0 || value <= 0.0 )
    {
        return false;
    }

    divisor_ = value;

    return true;
}

/**************************************************************************************************************/
bool
Stat2Report::setMinimalTraffic( double value )
{
    if ( value <= 0.0 )
    {
        return false;
    }

    minimalTraffic_ = value;

    return true;
}

/**************************************************************************************************************/
void
Stat2Report::printHeader() const
{
    if ( !shouldHideTime_ )
    {
        cout << setw(16) << "time";
    }
    else if ( !shouldShowVlans_ && !shouldShowHosts_ && !shouldShowFlows_ )
    {
        cout << "   ";
    }

    if ( shouldShowVlans_ )
    {
        cout << setw(6) << "vlan";
    }

    cout << "   ";

    if ( shouldShowHosts_ )
    {
        cout << left << setw(18) << "host" << right;
    }

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << "i.flows" << setw(12) << "i.packets";
    }

    cout << setw(15) << "i.traffic";

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << "o.flows" << setw(12) << "o.packets";
    }

    cout << setw(15) << "o.traffic" << endl;
}

/**************************************************************************************************************/
bool
Stat2Report::printRow( const Stat2Query::Key& stat, const SummaryIo& summary ) const
{
    uintmax_t   minimalBytes = static_cast<uintmax_t>( minimalTraffic_ * divisor_ );

    if ( minimalBytes && summary.nBytesIn() < minimalBytes && summary.nBytesOut() < minimalBytes )
    {
        return false;
    }

    if ( !shouldHideTime_ )
    {
        static time_t  timeCache = 0;
        static char    timeCacheString[ 32 ];

        if ( timeCache != stat.group )
        {
            struct tm  tm;

            timeCache = stat.group;
            localtime_r( &timeCache, &tm );
            strftime( timeCacheString, sizeof(timeCacheString) - 1, "%FT%H:%M", &tm );
        }

        cout << setw(16) << timeCacheString;
    }
    else if ( !shouldShowVlans_ && !shouldShowHosts_ && !shouldShowFlows_ )
    {
        cout << "   ";
    }

    if ( shouldShowVlans_ )
    {
        cout << setw(6) << stat.vlan;
    }

    cout << "   ";

    if ( shouldShowHosts_ )
    {
        cout << left << setw(18) << stat.mac.str() << right;
    }

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << summary.nFlowsIn() << setw(12) << summary.nPacketsIn();
    }

    cout << setw(15);

    if ( divisor_ == 1.0 )
    {
        cout << summary.nBytesIn();
    }
    else
    {
        cout << fixed << setprecision(2) << summary.nBytesIn() / divisor_;
    }

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << summary.nFlowsOut() << setw(12) << summary.nPacketsOut();
    }

    cout << setw(15);

    if ( divisor_ == 1.0 )
    {
        cout << summary.nBytesOut() << endl;
    }
    else
    {
        cout << fixed << setprecision(2) << summary.nBytesOut() / divisor_ << endl;
    }

    return true;
}

/**************************************************************************************************************/
void
Stat2Report::printFooter( string message, const SummaryIo& total ) const
{
    cout << setw(19 + shouldHideTime_*-16 + shouldShowVlans_*6 + shouldShowHosts_*18) << left << message << right;

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << total.nFlowsIn() << setw(12) << total.nPacketsIn();
    }

    cout << setw(15);

    if ( divisor_ == 1.0 )
    {
        cout << total.nBytesIn();
    }
    else
    {
        cout << fixed << setprecision(2) << total.nBytesIn() / divisor_;
    }

    if ( shouldShowFlows_ )
    {
        cout << setw(12) << total.nFlowsOut() << setw(12) << total.nPacketsOut();
    }

    cout << setw(15);

    if ( divisor_ == 1.0 )
    {
        cout << total.nBytesOut() << endl;
    }
    else
    {
        cout << fixed << setprecision(2) << total.nBytesOut() / divisor_ << endl;
    }
}
